/******************************************************************************
*       SOFA, Simulation Open-Framework Architecture, development version     *
*                (c) 2006-2016 INRIA, USTL, UJF, CNRS, MGH                    *
*                                                                             *
* This library is free software; you can redistribute it and/or modify it     *
* under the terms of the GNU Lesser General Public License as published by    *
* the Free Software Foundation; either version 2.1 of the License, or (at     *
* your option) any later version.                                             *
*                                                                             *
* This library is distributed in the hope that it will be useful, but WITHOUT *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License *
* for more details.                                                           *
*                                                                             *
* You should have received a copy of the GNU Lesser General Public License    *
* along with this library; if not, write to the Free Software Foundation,     *
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA.          *
*******************************************************************************
*                               SOFA :: Modules                               *
* Authors: Igor Peterlik                                                      *
*                                                                             *
* Contact information: contact@sofa-framework.org                             *
******************************************************************************/

#ifndef SOFA_COMPONENT_BEZIER_SEGMENT_TREE_H
#define SOFA_COMPONENT_BEZIER_SEGMENT_TREE_H

#include <initBezierSkeletonization.h>
#include "BezierSegment.h"

#include <sofa/defaulttype/VecTypes.h>
#include <sofa/defaulttype/RigidTypes.h>
#include <sofa/core/behavior/MechanicalState.h>
#include <sofa/core/objectmodel/BaseObject.h>
#include <sofa/core/objectmodel/BaseContext.h>
#include <sofa/core/topology/BaseMeshTopology.h>
#include <sofa/simulation/AnimateBeginEvent.h>
#include <sofa/core/loader/BaseLoader.h>
#include <sofa/core/loader/PrimitiveGroup.h>
#include <sofa/core/topology/Topology.h>
#include <sofa/helper/fixed_array.h>

namespace sofa
{

namespace component
{

namespace misc
{

using namespace sofa::defaulttype;
using namespace sofa::helper;
using namespace sofa::bezier4D;

template<class DataTypes>
class BezierSegmentTree : public virtual sofa::core::objectmodel::BaseObject {
public:
    SOFA_CLASS( SOFA_TEMPLATE(BezierSegmentTree, DataTypes) , sofa::core::objectmodel::BaseObject);

    typedef typename sofa::defaulttype::Vec3dTypes::VecCoord VecCoord3D;
    typedef typename sofa::defaulttype::Vec3dTypes::Coord Coord3D;
    typedef typename DataTypes::VecCoord VecCoord;
    typedef typename DataTypes::VecDeriv VecDeriv;
    typedef typename DataTypes::Coord Coord;
    typedef typename DataTypes::Deriv Deriv;
    typedef typename Coord::Pos Pos;
    typedef typename Coord::Quat Quat;
    typedef typename Coord::Real Real;
    typedef defaulttype::Vec3d Vec3;

    typedef std::vector<CurvePoint4> CurveVec;

    typedef sofa::core::topology::Topology::Edge Edge;
    typedef std::pair<int,int> IPair;

    BezierSegmentTree();
    ~BezierSegmentTree();
    void init();
    void bwdInit();

    Data<string> d_treeFile;
    Data<Real> d_distanceBetweenPoints;
    Data<Real> d_collisionOffset;
    Data<VecCoord> d_positions;    
    Data<VecCoord3D> d_collisionPositions;
    Data<vector<Edge> > d_edges;
    Data<vector<Edge> > d_collisionEdges;
    Data<vector<int> > d_inputPointDataSegmentIDs;
    Data<vector<int> > d_inputTopology;
    Data<vector<Pos> > d_inputPositions;
    Data<vector<double> > d_pointRadii;
    Data<double> d_fittingError;
    Data<double> d_defaultRadius;    
    Data<int> d_maxSegments;

    Data<int> d_showCurveDiscretization;
    Data<double> d_showInputPointSize;
    Data<double> d_showControlPointSize;
    Data<double> d_showTangentSize;
    Data<double> d_showSampledPointsSize;
    Data<double> d_showCurve;
    Data<std::string> d_xmlFilePrefix;
    Data<std::string> d_swcFilePrefix;
    Data<bool> d_verbose;
            
    Data<float> d_cnodesIDFontSize;
    Data<double> d_drawOffset;
    Data<std::string> d_treeCorrespondFile;
    
    Data<VecCoord3D> d_matchingPositions1;
    Data<VecCoord3D> d_matchingPositions2;
    Data<vector<Edge> > d_matchingEdges;
    Data<vector<int> > d_matchingEdgesIDs;
    Data<vector<int> > d_matchingPointsIDs;
    Data<VecCoord> d_treeRoots;
    Data<double>    d_truncateOpenSegments;
    Data<vector<int> > d_bifurcationParentNodes;
    Data<vector<int> > d_bifurcationChildNodes;
    Data<VecCoord3D> d_bifurcationPositions;
    Data<bool> d_subdivideInputSegments;
    Data<bool> d_constructBifurcationPairs;
    Data<Vec3> d_treeColor;
    
    SingleLink<BezierSegmentTree<DataTypes>, BezierSegmentTree<DataTypes>, BaseLink::FLAG_STOREPATH|BaseLink::FLAG_STRONGLINK> l_matchingBezierTreeLink;

    CurveVec cnodes;
    int totalNumVessels, totalNumSegments;
    
    BezierSegment* findSegmentFromIDs(int ID1, int ID2);
    BezierSegment* findSegmentFromPredCnode(int predCnode);
    int getPredIdxSWC(BezierSegment* predSeg, RigidCurvePoint4 predCnode);
    std::map<IPair, IPair> segmentIDsToIndices;
    vector<std::string> segQualities;
    vector<double> segTruncate;

    std::vector<BezierSegment>& getBezierSegments() {
        return bezierSegments;
    }

protected:    
    std::vector<BezierSegment> bezierSegments;
    
    /// variables related to tree matching
    vector<vector<int> >  matchingSegmentsIDs[2];
    
    std::vector<BezierSegment> matchSegments1, matchSegments2;
        
    void readTreeCorrespondences();
    void matchBezierSegments();
    void matchSegmentsByRefitting();
    int readCNodesFromCSVFiles();
    int createCNodesFromTopology();

    int selectBranchingCNodes();

    int constructSegmentsFromCNodes();

    bool createBifurcationIndices();

    void printCNodes();


    void saveToXML(string filePrefix);
    void saveToSWC(string filePrefix);

    //void constructSampledPoints(Real dist);
    void constructSampledPoints();

public:

    virtual std::string getTemplateName() const
    {
      return templateName(this);
    }

    static std::string templateName(const BezierSegmentTree<DataTypes>* = NULL)
    {
      return DataTypes::Name();
    }

    /// Pre-construction check method called by ObjectFactory.
    /// Check that DataTypes matches the MechanicalState.
    template<class T>
    static bool canCreate(T* obj, sofa::core::objectmodel::BaseContext* context, sofa::core::objectmodel::BaseObjectDescription* arg)
    {
            //if (dynamic_cast<sofa::core::behavior::MechanicalState<DataTypes>*>(context->getMechanicalState()) == NULL)
            //{
            //        return false;
            //}
            return BaseObject::canCreate(obj, context, arg);
    }

    void draw(const core::visual::VisualParams* vparams);

    Pos getPos(Point4& _p) {
        Pos x;
        for (size_t i = 0; i < 3; i++)
            x[i] = _p[i];
        return(x);
    }

    Pos getPos(Point3& _p) {
        Pos x;
        for (size_t i = 0; i < 3; i++)
            x[i] = _p[i];
        return(x);
    }

private:
    vector<double> rcol;
    vector<double> gcol;
    vector<double> bcol;

};


#if defined(SOFA_EXTERN_TEMPLATE) && !defined(SOFA_COMPONENT_BEZIER_SEGMENT_TREE_CPP)
#ifndef SOFA_FLOAT
extern template class SOFA_MISC_DEV_API BezierSegmentTree<defaulttype::Rigid3dTypes>;
#endif
#ifndef SOFA_DOUBLE
extern template class SOFA_MISC_DEV_API BezierSegmentTree<defaulttype::Rigid3fTypes>;
#endif
#endif

}

}

}

#endif
