/******************************************************************************
*       SOFA, Simulation Open-Framework Architecture, development version     *
*                (c) 2006-2016 INRIA, USTL, UJF, CNRS, MGH                    *
*                                                                             *
* This library is free software; you can redistribute it and/or modify it     *
* under the terms of the GNU Lesser General Public License as published by    *
* the Free Software Foundation; either version 2.1 of the License, or (at     *
* your option) any later version.                                             *
*                                                                             *
* This library is distributed in the hope that it will be useful, but WITHOUT *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License *
* for more details.                                                           *
*                                                                             *
* You should have received a copy of the GNU Lesser General Public License    *
* along with this library; if not, write to the Free Software Foundation,     *
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA.          *
*******************************************************************************
*                               SOFA :: Modules                               *
*                                                                             *
* Authors: The SOFA Team and external contributors (see Authors.txt)          *
*                                                                             *
* Contact information: contact@sofa-framework.org                             *
******************************************************************************/
#ifndef SOFA_CORE_COLLISION_PARDISOCONSTRAINTCORRECTION_INL
#define SOFA_CORE_COLLISION_PARDISOCONSTRAINTCORRECTION_INL

#include "PardisoConstraintCorrection.h"
#include <sofa/defaulttype/RigidTypes.h>
#include <sofa/simulation/Node.h>
#include <sofa/simulation/MechanicalVisitor.h>
#include <sofa/core/visual/VisualParams.h>
#include <sofa/core/behavior/ConstraintCorrection.inl>

#include <sstream>
#include <list>

namespace sofa
{

namespace component
{

namespace constraintset
{

template<class DataTypes>
PardisoConstraintCorrection<DataTypes>::PardisoConstraintCorrection(sofa::core::behavior::MechanicalState<DataTypes> *mm)
    : Inherit(mm)
    , d_schurSolverName( initData(&d_schurSolverName, "schurSolverName", "name of solver that computes the Schur complement (optimized CPU version with Pardiso only)") )
    , d_showTiming( initData(&d_showTiming,false,"showTiming","if true, show length of computations in microseconds") )
    , d_filteringThreshold( initData(&d_filteringThreshold, double(0.0), "filteringThreshold", "filter all values in matrix J lower than this threshold, inactive by default") )
{

}


template<class DataTypes>
void PardisoConstraintCorrection<DataTypes>::init()
{
    Inherit::init();

    const helper::vector<std::string>& schurSolverNames = d_schurSolverName.getValue();

    schurLinearSolvers.clear();
    if (schurSolverNames.size() > 0)
    {
        sofa::core::objectmodel::BaseContext* c = this->getContext();
        for (unsigned int i=0; i<schurSolverNames.size(); ++i) {
            sofa::core::behavior::LinearSolver* s = NULL;
            c->get(s, schurSolverNames[i]);
            if (s) {
                schurLinearSolvers.push_back(s);
                if (this->f_printLog.getValue())
                    std::cout << "Found Schur solver: " << s->getName() << std::endl;
            }
        }
    }

    std::cout << "Schur solvers: " << schurLinearSolvers.size() << std::endl;
    if (schurLinearSolvers.size() != this->linearsolvers.size())
        schurLinearSolvers = this->linearsolvers;
}

template<class DataTypes>
void PardisoConstraintCorrection<DataTypes>::addComplianceInConstraintSpace(const sofa::core::ConstraintParams *cparams, sofa::defaulttype::BaseMatrix* W)
{
    if (!this->mstate || !this->odesolver || (this->linearsolvers.size()==0)) return;

    // use the OdeSolver to get the position integration factor
    double factor = 1.0;

    switch (cparams->constOrder())
    {
    case core::ConstraintParams::POS_AND_VEL :
    case core::ConstraintParams::POS :
        factor = this->odesolver->getPositionIntegrationFactor();
        break;

    case core::ConstraintParams::ACC :
    case core::ConstraintParams::VEL :
        factor = this->odesolver->getVelocityIntegrationFactor();
        break;

    default :
        break;
    }

    const unsigned int numDOFs = this->mstate->getSize();
    const unsigned int N = Deriv::size();
    const unsigned int numDOFReals = numDOFs*N;

    // Compute J
    const MatrixDeriv& c = this->mstate->read(core::ConstMatrixDerivId::holonomicC())->getValue();
    const unsigned int totalNumConstraints = W->rowSize();

    this->J.resize(totalNumConstraints, numDOFReals);

    MatrixDerivRowConstIterator rowItEnd = c.end();

    double filtThresh = d_filteringThreshold.getValue();

    for (MatrixDerivRowConstIterator rowIt = c.begin(); rowIt != rowItEnd; ++rowIt)
    {
        const int cid = rowIt.index();

        MatrixDerivColConstIterator colItEnd = rowIt.end();

        for (MatrixDerivColConstIterator colIt = rowIt.begin(); colIt != colItEnd; ++colIt)
        {
            const unsigned int dof = colIt.index();
            const Deriv n = colIt.val();            

            for (unsigned int r = 0; r < N; ++r)
            {
                if (fabs(n[r]) > filtThresh)
                    this->J.add(cid, dof * N + r, n[r]);
            }
        }
    }

    //std::cout << "[" <<  this->getName() << "] J = " << this->J << std::endl;

    // use the Linear solver to compute J*inv(M)*Jt, where M is the mechanical linear system matrix
    for (unsigned i = 0; i < schurLinearSolvers.size(); i++)
    {
        double startTime, stopTime;
        sofa::helper::system::thread::CTime *timer;

        if (d_showTiming.getValue())
            startTime = double(timer->getTime());

        this->linearsolvers[i]->setSystemLHVector(sofa::core::MultiVecDerivId::null());
        schurLinearSolvers[i]->addJMInvJt(W, &this->J, factor);

        if (d_showTiming.getValue()) {
            stopTime = double(timer->getTime());
            std::cout << "[" << this->getName() << "] WTIME: compute Schur total" << " " << stopTime - startTime << std::endl;
        }

    }

}

}   // constraintset namespace

}   // component

}   // sofa

#endif
